package de.mbutscher.wikiandpad;

import de.mbutscher.wikiandpad.utils.MiscEvent;
import de.mbutscher.wikiandpad.utils.MiscEventSource;
import de.mbutscher.wikiandpad.utils.ui.FastScroller;
import de.mbutscher.wikiandpad.utils.ui.FastScroller.OnScrollListener;
import de.mbutscher.wikiandpad.utils.ui.Touch;
import android.content.Context;
import android.content.SharedPreferences;
import android.graphics.Canvas;
import android.text.Layout;
import android.text.Selection;
import android.util.AttributeSet;
import android.view.GestureDetector;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.view.inputmethod.InputMethodManager;
import android.widget.EditText;


/**
 * EditText variant with support for vertical fling scrolling.
 * Based on http://stackoverflow.com/questions/6422697/vertical-fling-scrolling-of-text-line-in-android
 * TODO: Make smoother
 */
public class ScrollingEditText extends EditText implements MiscEventSource {

    /** Constants for {@link #setSelection(int, int, ScrollTo)}
     */
    public enum ScrollTo {
        /** No special scrolling
         */
        DEFAULT,

        /** Scroll selection to top
         */
        TOP,

        /** Scroll selection to vertical middle
         */
        MIDDLE
    }

    protected boolean didMove;
    protected boolean workaround_tabKey;
    protected GestureDetector gestureDetector;
    protected FastScroller mFastScroller;
    
    protected MiscEvent miscEvent = new MiscEvent(this);

    
    protected GestureDetector.SimpleOnGestureListener dblClickListener =
            new GestureDetector.SimpleOnGestureListener() {

        @Override
        public boolean onDoubleTap(MotionEvent e) {
            
            Layout layout = ScrollingEditText.this.getLayout();
            final int selStart;
            if (layout != null)
            {
                int line = layout.getLineForVertical((int) e.getY());
                selStart = layout.getOffsetForHorizontal(line, (int) e.getX());
            }
            else
                return false;
            
            return activateSelection(selStart);
        }
    };
    public ScrollingEditText(Context context, AttributeSet attrs, int defStyle) {
        super(context, attrs, defStyle);
        init(context);
    }

    public ScrollingEditText(Context context, AttributeSet attrs) {
        super(context, attrs);
        init(context);
    }

    public ScrollingEditText(Context context) {
        super(context);
        init(context);
    }
    
    public MiscEvent getMiscEvent() {
        return miscEvent;
    }


    protected void init(Context context) {
        if (!isInEditMode()) {
            SharedPreferences pref = WikiAndPadApplication.getAppPrefs();
            workaround_tabKey = pref.getBoolean("workaround_tabKey", false);
        }

        mFastScroller = new FastScroller(getContext(), this);
        
        mFastScroller.setOnScrollListener(new OnScrollListener() {
            
            public void onScroll(View view, float yPosition) {
                int line = (int) (yPosition * getLineCount());
                moveToLine(line);
            }
        });
    }

    @Override
    protected void onDraw(Canvas canvas) {
        super.onDraw(canvas);
        if (mFastScroller != null) {
            mFastScroller.draw(canvas);
        }
        
    }
    
    public void setFastScrollEnabled(boolean enabled) {
        if (enabled) {
            if (mFastScroller == null) {
                mFastScroller = new FastScroller(getContext(), this);
            }
        } else {
            if (mFastScroller != null) {
                mFastScroller.stop();
                mFastScroller = null;
            }
        }
    }

    
    @Override
    public boolean onTouchEvent(MotionEvent event) {
        if (mFastScroller != null) {
            boolean intercepted = mFastScroller.onInterceptTouchEvent(event);
            if (intercepted) {
                Touch.cancelFling(this, getText());
                return true;
            }
            intercepted = mFastScroller.onTouchEvent(event);
            if (intercepted) {
                Touch.cancelFling(this, getText());
                return true;
            }
        }
        
        if (Touch.onTouchEvent(this, getText(), event))
            return true;

        return super.onTouchEvent(event);
    }
    
    @Override
    protected void onSizeChanged(int w, int h, int oldw, int oldh) {
        if (mFastScroller != null) {
            mFastScroller.onSizeChanged(w, h, oldw, oldh);
        }
        super.onSizeChanged(w, h, oldw, oldh);
    }
    
    public int getVisibleHeight()
    {
        return getBottom() - getTop() - getExtendedPaddingBottom() - getExtendedPaddingTop();
    }

    
    @Override
    protected void onScrollChanged(int l, int t, int oldl, int oldt) {
        super.onScrollChanged(l, t, oldl, oldt);

        if (mFastScroller != null) {
            mFastScroller.onScroll(this, t, getVisibleHeight(),
                    getLayout().getHeight());
        }
    }
    
    @Override
    public boolean isVerticalScrollBarEnabled() {
        return !(mFastScroller != null && mFastScroller.isVisible());
    }
    
    
    /** {@link #getSelectionStart()} can be greater than
     * {@link #getSelectionEnd()} which can lead to a crash in some
     * functions
     * @return The smaller result of the above two functions 
     */
    public int getSelectionLow() {
        return Math.min(getSelectionStart(), getSelectionEnd());
    }
    
    /** {@link #getSelectionStart()} can be greater than
     * {@link #getSelectionEnd()} which can lead to a crash in some
     * functions
     * @return The greater result of the above two functions 
     */
    public int getSelectionHigh() {
        return Math.max(getSelectionStart(), getSelectionEnd());
    }

    public void moveToLine(int line){
        int offset = getLayout().getLineStart(line);
        Selection.setSelection(getText(), offset,  offset );
    }


    
    public void hideSoftInput() {
        InputMethodManager imm = (InputMethodManager)
                getContext().getSystemService(Context.INPUT_METHOD_SERVICE);
        
        imm.hideSoftInputFromWindow(getWindowToken(), 0);
    }
    
    public boolean hasSoftInput() {
        InputMethodManager imm = (InputMethodManager)
                getContext().getSystemService(Context.INPUT_METHOD_SERVICE);
        
        return imm.isActive();
    }
    
    
    public boolean activateSelection(int selStart) {
        if (selStart == -1)
            return false;


        boolean processed = getMiscEvent().cloneIt().put("activated text")
                .put("textOffset", selStart).send()
                .hasKey("processed");

        return processed;
    }
    
    public boolean activateSelection() {
        return activateSelection(getSelectionStart());
    }
    
    @Override
    public boolean dispatchKeyEvent(KeyEvent event) {
        // Log.d("WikiAndPad", "dispatchKeyEvent " + event.getAction());
        if (workaround_tabKey && event.getUnicodeChar() == 9) {
            if (event.getAction() == KeyEvent.ACTION_DOWN)
                getText().replace(getSelectionLow(), getSelectionHigh(), "\u0009");

            return true;
        }

        return super.dispatchKeyEvent(event);
    }
    
    
    /**
     * Set selection
     * @param start
     * @param end
     * @param scrollTo How to scroll after selecting
     */
    public void setSelection(int start, int end, ScrollTo scrollTo) {
        setSelection(start, end);
        
        switch (scrollTo) {
        case TOP: {
            Layout layout = getLayout();
            if (layout != null) {
                int line = layout.getLineForOffset(start);
                int yPos = layout.getLineBaseline(line)
                        + layout.getLineAscent(line);
                scrollTo(getScrollX(), yPos);
            }
            break;
        }
        case MIDDLE: {
            Layout layout = getLayout();
            if (layout != null) {
                int line = layout.getLineForOffset(start);
                int yPos = layout.getLineBaseline(line)
                        + layout.getLineAscent(line);
                int innerHeight = getHeight() -
                        getPaddingTop() - getPaddingBottom();
                
                scrollTo(getScrollX(), yPos - innerHeight / 2);
            }
            break;
        }
        default:
            break;
        }
    }


    

    

}
