package de.mbutscher.wikiandpad;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import de.mbutscher.wikiandpad.utils.HashMapBuilder;
import de.mbutscher.wikiandpad.utils.StringOps;

public class WikiConfig {

    protected Map<String, String> configDefault;
    protected Map<String, String> configData;
    protected String configPath;
    protected boolean writeAccessDenied = false;
    
    public WikiConfig(Map<String, String> configDefault) {
        this.configDefault = configDefault;
    }


    public static Map<String, String> parseIni(String iniData) {
        HashMap<String, String> result = new HashMap<String, String>();
        for(String line: iniData.split("\\r\\n?|\\n")) {
            String[] kv = line.split("=", 2);
            if (kv.length != 2)
                continue;
            kv[0] = kv[0].trim();
            kv[1] = kv[1].trim();

            result.put(kv[0], kv[1]);
        }
        
        return result;
    }


    /**
     * The function checks if an option is valid.
        Only options can be set/retrieved which have an entry in the
        defaults and if the configParserObject is valid.
     * @param option
     * @return
     */
    public boolean isOptionAllowed(String option) {
        return (configData != null && configDefault.containsKey(option));
    }


    
    
    public void testOptionAllowed(String option) {
        if (!isOptionAllowed(option))
            throw new UnknownOptionException("Unknown option %s", option);
    }


    /**
     * Return a configuration value returned as unicode which
        has specified option key.
     * @param option
     * @param defValue
     * @return
     */
    public String get(String option, String defValue) {
        testOptionAllowed(option);
        
        if (configData.containsKey(option))
            return configData.get(option);
        
        if (configDefault.containsKey(option))
            return configDefault.get(option);
        
        return defValue;
    }

    public String get(String option) {
        return get(option, null);
    }
    
    

    public int getInt(String option, int defValue) {
        String v = get(option, null);
        
        if (v == null)
            return defValue;
        
        try {
            return Integer.parseInt(v);
        }
        catch (NumberFormatException e) {
            return defValue;
        }
    }
    
    
    public double getDouble(String option, double defValue) {
        String v = get(option, null);
        
        if (v == null)
            return defValue;
        
        try {
            return Double.parseDouble(v);
        }
        catch (NumberFormatException e) {
            return defValue;
        }
    }

    
    public boolean getBoolean(String option, boolean defValue) {
        String v = get(option, null);
        
        if (v == null)
            return defValue;
        
        return StringOps.strToBool(v , defValue);
    }
    
    
    // TODO Allow in read-only mode?
    public WikiConfig put(String option, String value) {
        testOptionAllowed(option);
        configData.put(option, value);
        return this;
    }
    
    public WikiConfig put(String option, int value) {
        testOptionAllowed(option);
        put(option, Integer.toString(value));
        return this;
    }

    public WikiConfig put(String option, double value) {
        testOptionAllowed(option);
        put(option, Double.toString(value));
        return this;
    }

    public WikiConfig put(String option, boolean value) {
        testOptionAllowed(option);
        if (value)
            put(option, "True");
        else
            put(option, "False");
        return this;
    }

    
    public void loadConfig(String path)
            throws FileNotFoundException, IOException {
        if (path == null)
            configData = new HashMap<String, String> ();
        
        String data = StringOps.loadEntireTxtFile(path);
        configPath = path;
        configData = parseIni(data);
    }
    
    
    public void saveConfig() throws IOException {
        StringBuilder result = new StringBuilder();
        
        result.append("[main]\n");
        
        // if self.isReadOnlyEffect(): return
        if (configPath != null && configData != null) {
            for(Map.Entry<String, String> entry: configData.entrySet()) {
                if (entry.getKey().contains("="))
                    continue;
                result.append(entry.getKey() + "=" + entry.getValue() + "\n");
            }
        }
        StringOps.writeEntireFile(configPath, result.toString(), true);
    }

    
    
    
    public static final Map<String, String> WIKI_CONFIG_DEFAULTS =
            (new HashMapBuilder<String, String>())
            .put("data_dir", "data")
            .put("wiki_name", "")
            .put("wiki_wikiLanguage", "wikidpad_default_2_0")
            .put("wiki_database_type", "")
            .getHashMap();

}


/*






*/