package de.mbutscher.wikiandpad;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.Collator;
import java.util.ArrayList;
import java.util.List;
import java.util.WeakHashMap;

import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.net.Uri;

import de.mbutscher.wikiandpad.db.DbReadAccessException;
import de.mbutscher.wikiandpad.db.DbWriteAccessException;
import de.mbutscher.wikiandpad.db.IWikiData;
import de.mbutscher.wikiandpad.db.MatchTermEntry;
import de.mbutscher.wikiandpad.db.WikiDataException;
import de.mbutscher.wikiandpad.utils.FileUtils;
import de.mbutscher.wikiandpad.utils.NotCurrentThreadException;
import de.mbutscher.wikiandpad.utils.StringOps;
import de.mbutscher.wikiandpad.utils.Utilities;

public class WikiDocument {

    protected static WikiDocument activeDocument;
    protected static boolean startupTryDone = false; 
    
    /**
     * Each time database changes it is incremented. If an Activity
     * wants to know if a change happened it stores the current count
     * and compares it later.
     */
    protected long databaseChangeCount;
    
    protected String wikiName;
    // protected String wikiDataDir;
    protected IWikiData wikiData;
    
    protected WeakHashMap<String, WikiPage> docPagePool =
            new WeakHashMap<String, WikiPage>();
    
    protected WikiConfig wikiConfig;

    /** You should not use the constructor directly, use
     * {@link #openDocument(String)} instead.
     * @param wikiConfigPath
     * @throws IOException 
     * @throws FileNotFoundException 
     * @throws WikiDataException 
     */
    protected WikiDocument(String wikiConfigPath)
            throws FileNotFoundException, IOException, WikiDataException {
        wikiConfig = new WikiConfig(WikiConfig.WIKI_CONFIG_DEFAULTS);
        
        try {
            wikiConfig.loadConfig(wikiConfigPath);
        }
        catch(OutOfMemoryError e) {
            throw new WikiDataException(
                    "Configuration file '%s' too large (has %s bytes). Maybe " +
                    "not a real config file.",
                    wikiConfigPath, Long.toString(
                    (new File(wikiConfigPath)).length()));
        }
        
        if ("original_sqlite".equals(wikiConfig.get("wiki_database_type")))
            wikiData = new de.mbutscher.wikiandpad.db.originalSqlite.WikiData(
                    wikiConfigPath, this);
        else if ("compact_sqlite".equals(wikiConfig.get("wiki_database_type")))
            wikiData = new de.mbutscher.wikiandpad.db.compactSqlite.WikiData(
                    wikiConfigPath, this);
        else {
            throw new WikiDataException(
                    "Only database types \"Original Sqlite\" and " + 
                    "\"Compact Sqlite\" are supported");
        }

        wikiName = (new File(wikiConfigPath)).getName();
        int lastDot = wikiName.lastIndexOf(".");
        if (lastDot > -1)
            wikiName = wikiName.substring(0, lastDot);
    }
    
    public String getWikiName() {
        return wikiName;
    }

    public IWikiData getWikiData() {
        return wikiData;
    }
    
    public long getDatabaseChangeCount() {
        return databaseChangeCount;
    }
    
    public void newDatabaseChangeCount() {
        ++databaseChangeCount;
    }

    public synchronized boolean isSameDatabaseChangeCount(long other) {
        return other == databaseChangeCount;
    }

    
    public synchronized String getWikiConfigPath() {
        if (wikiData == null)
            return null;
        return wikiData.getWikiConfigPath();
    }

    public synchronized File getWikiConfigDirFile() {
        if (wikiData == null)
            return null;
        
        return (new File(getWikiConfigPath())).getParentFile();
    }

    public synchronized String getWikiConfigDir() {
        if (wikiData == null)
            return null;
        
        File fTemp = (new File(getWikiConfigPath())).getParentFile();
        return fTemp.getAbsolutePath();
    }

    public synchronized void close() {
        if (wikiData != null) {
            wikiData.close();
            wikiData = null;
        }
    }
    
    public boolean isReadOnlyEffect() {
        // TODO
        return false;
    }

    public static synchronized WikiDocument getActiveDocument() {
        if (!startupTryDone) {
            startupTryDone = true;
            SharedPreferences pref = WikiAndPadApplication.getAppPrefs();
            String lastWikiPath = pref.getString("last_wiki", null);
            if (lastWikiPath != null) {
                try {
                    openDocument(lastWikiPath, false);
                } catch (FileNotFoundException e) {
                    activeDocument = null;
                } catch (IOException e) {
                    activeDocument = null;
                } catch (WikiDataException e) {
                    activeDocument = null;
                }
            }
        }
        return activeDocument;
    }

    public static synchronized void setActiveDocument(
            WikiDocument activeDocument, boolean storeInPrefs) {
        if (WikiDocument.activeDocument != null)
            WikiDocument.activeDocument.close();

        WikiDocument.activeDocument = activeDocument;
        startupTryDone = true;
        
        if (storeInPrefs && activeDocument != null) {
            Editor prefEd = WikiAndPadApplication.getAppPrefs().edit();
            prefEd.putString("last_wiki", activeDocument.getWikiConfigPath());
            prefEd.commit();
        }
    }

    public static void setActiveDocument(WikiDocument activeDocument) {
        setActiveDocument(activeDocument, true);
    }

        
    public static synchronized WikiDocument openDocument(String wikiConfigPath,
            boolean storeInPrefs)
            throws FileNotFoundException, IOException, WikiDataException {
        
        if (wikiConfigPath == null) {
            if (storeInPrefs) {

                Editor prefEd = WikiAndPadApplication.getAppPrefs().edit();
                prefEd.remove("last_wiki");
                setActiveDocument(null, false);
                prefEd.commit();
            }
            return activeDocument;
        }
        
        WikiDocument active = getActiveDocument();
        if (active == null || !active.getWikiConfigPath().equals(wikiConfigPath)) {
            // Only create new document if not same as active one
            setActiveDocument(new WikiDocument(wikiConfigPath), storeInPrefs);
        }

        return activeDocument;
    }

    public static WikiDocument openDocument(String wikiConfigPath)
            throws FileNotFoundException, IOException, WikiDataException {
        return openDocument(wikiConfigPath, true);
    }

    
    /**
     * get the list of match terms with thisStr in them.
     * @param thisStr
     * @param orderBy
     * @param descend
     * @return
     */
    public List<MatchTermEntry> getWikiWordMatchTermsWith(String thisStr,
            String orderBy, boolean descend) {
        return wikiData.getWikiWordMatchTermsWith(thisStr, orderBy, descend);
    }
    
    public List<MatchTermEntry> searchWiki(SearchAndReplaceOperation sarOp)
            throws DbReadAccessException, NotCurrentThreadException {
        ArrayList<MatchTermEntry> result = new ArrayList<MatchTermEntry>();

        if (sarOp == null)
            return result;

        for(String wpn = wikiData.getFirstWikiPageName(); wpn != null;
                wpn = wikiData.getNextWikiPageName(wpn)) {

            Utilities.testCurrentValidThread();
            String content = getWikiPageContent(wpn);

            MatchTermEntry mte = sarOp.searchText(content, 0);
            if (mte != null) {
                mte.wikiPageName = wpn;
                result.add(mte);
            }
        }

        return result;
    }


    public Collator getCollator() {
        return Collator.getInstance();
    }
    
    public WikiPage getWikiPage(String wikiPageName) {
        String unifName = "wikipage/" + wikiPageName;
        
        WikiPage result = docPagePool.get(unifName);
        if (result == null) {
            result = new WikiPage(this, wikiPageName);
            docPagePool.put(unifName, result);
        }

        return result;
    }

    public String getWikiPageContent(String wikiPageName)
            throws DbReadAccessException {

        return wikiData.getWikiPageContent(wikiPageName);
    }

    public String getUnifiedPageContent(String unifName)
            throws DbReadAccessException {
        if (!unifName.startsWith("wikipage/"))
            return null;

        return getWikiPageContent(unifName.substring(9));
    }

    public void setWikiPageContent(String wikiPageName, String content,
            double moddate, double creadate) throws DbWriteAccessException {
        wikiData.setWikiPageContent(wikiPageName, content, moddate, creadate);
    }
    
    public void setWikiPageContent(String wikiPageName, String content)
            throws DbWriteAccessException {
        setWikiPageContent(wikiPageName, content, -1, -1);
    }

    public void updateWikiWordMatchTerms(String wikiPageName,
            List<MatchTermEntry> wwmTerms, boolean syncUpdate) {
        wikiData.updateWikiWordMatchTerms(wikiPageName, wwmTerms, syncUpdate);
    }

    /**
     * 
     * @param wikiPageName
     * @param toWord
     * @param modifyText Should the text of links to the renamed page be
                modified? This text replacement works unreliably
     * @throws WikiDataException
    */
    public void renameWikiWord(String wikiPageName, String toWord,
            boolean modifyText)
            throws DbWriteAccessException, WikiDataException {
        // TODO: Check for invalid wiki word
        // TODO: Check for all link terms
        if (wikiData.isDefinedWikiPageName(toWord))
            throw new WikiDataException(String.format("Rename target %s " +
                    "exists already", toWord));
        
        wikiData.renameWord(wikiPageName, toWord);
        WikiPage oldPage = getWikiPage(wikiPageName);
        oldPage.renameVersionData(toWord);
        oldPage.invalidate();

        WikiPage newPage = getWikiPage(toWord);
        newPage.refreshSyncUpdateMatchTerms();
    }
    
    
    public boolean isCreatableWikiWord(String wikiWord) {
        return !StringOps.notEmpty(getWikiPageNameForLinkTerm(wikiWord));
    }
    

    /**
     * Resolve links to wiki words. Returns null if it can't be resolved
     * @param linkTerm
     * @return
     */
    public String getWikiPageNameForLinkTerm(String linkTerm) {
        // TODO: Resolve properly in caseless mode
        return getWikiData().getWikiPageNameForLinkTerm(linkTerm);
    }
    
    public boolean isDefinedWikiPageName(String wikiPageName) {
        return getWikiData().isDefinedWikiPageName(wikiPageName);
    }
    
    
    /**
     * return the real word if word is an alias.
        returns word itself if word isn't an alias (may mean it's a real word
                or doesn't exist!)
     * @return
     */
    public String getWikiPageNameForLinkTermOrAsIs(String linkTerm) {
        String result = getWikiData().getWikiPageNameForLinkTerm(linkTerm);

        if (result == null)
            return linkTerm;

        return result;
    }
    
    
    /**
     * Convert a relative or absolute file URL to an absolute path.
        Returns None if fileurl isn't "file:" or "rel:" protocol
     * @param fileurl
     * @return
     */
    public File makeFileUrlAbsFile(String fileurl) {

        Uri uri = Uri.parse(fileurl);
        
        if (fileurl.startsWith("rel://")) {
            String relpath = Uri.decode(fileurl.substring(6));
            File f = new File(FileUtils.join(getWikiConfigDir(), relpath));
            
            return f.getAbsoluteFile();
        }
        else if (fileurl.startsWith("file:"))
            return new File(uri.getPath());
        else
            return null;
    }
    
    
    /**
     * Return the absolute file: URL for a rel: URL
     * @param relurl
     * @return
     */
    public String makeRelUrlAbsolute(String relurl) {
        
        if (relurl.startsWith("rel://")) {
            String relpath = Uri.decode(relurl.substring(6));
            String absPath = FileUtils.join(getWikiConfigDir(), relpath);
            
            /*
            File f = new File(getWikiConfigDir(), relpath);
            f = new File(f.getAbsolutePath());
            */
            return "file://" + StringOps.urlQuote(absPath, "/");

            // return Uri.fromFile(f).toString();
        }
        else if (relurl.startsWith("wikirel://")) {
            String relpath = Uri.decode(relurl.substring(10));
            String absPath = FileUtils.join(getWikiConfigDir(), relpath);
            
            return "wiki://" + StringOps.urlQuote(absPath, "/");

            /*
            File f = new File(getWikiConfigDir(), relpath);
            f = f.getAbsoluteFile();

            String url = "wiki:" + Uri.fromFile(f).toString().substring(5);

            return url;
            
            */
            
            
        }
        
        return null;
    }
    
    
    /**
     * Return the rel: URL for an absolute file path or None if
        a relative URL can't be created
     * @param absPath
     * @return
     */
    public String makeAbsPathRelUrl(String absPath) {
        String locPath = getWikiConfigDir();

        if (locPath == null)
            return null;

        String relPath = FileUtils.getRelativeFilePathIfPossible(locPath,
                absPath);

        if (relPath == null)
            return null;

        // TODO: Check if Uri.encode encodes '/'
        return "rel://" + StringOps.urlQuote(relPath);
    }

}


/*


    def makeAbsPathRelUrl(self, absPath, addSafe=''):
        """
        Return the rel: URL for an absolute file path or None if
        a relative URL can't be created
        """
        locPath = self.getWikiConfigPath()

        if locPath is None:
            return None

        locPath = os.path.dirname(locPath)
        relPath = relativeFilePath(locPath, absPath)
        if relPath is None:
            return None

        return u"rel://" + urlFromPathname(relPath, addSafe=addSafe)



*/


