package de.mbutscher.wikiandpad.db;

import de.mbutscher.wikiandpad.utils.Utilities;
import android.database.AbstractWindowedCursor;
import android.database.sqlite.SQLiteDatabase;

public class ConnectWrapSqlite {
    
    protected SQLiteDatabase database;
    protected int innerBogusTransactionCount = 0;
    

    public ConnectWrapSqlite(SQLiteDatabase database)
    {
        this.database = database;
    }
    
    /**
     * This variant is suitable for API level 8 to 10. For level 11 and above
     * there is a getType() method in Cursor.
     * @param cursor
     * @param column
     * @return Appropriate object depending on type of field
     */
    public static Object getObjectForItem(AbstractWindowedCursor cursor, int column) {
        if (cursor.isNull(column))
            return null;
        else if (cursor.isLong(column))
            return cursor.getLong(column);
        else if (cursor.isFloat(column))
            return cursor.getDouble(column);
        else if (cursor.isString(column))
            return cursor.getString(column);
        else if (cursor.isBlob(column))
            return cursor.getBlob(column);
        
        return null;
    }
    
    public static String escapeForGlob(String s) {
        StringBuilder result = new StringBuilder();
        
        char c;
        
        for(int i = 0; i < s.length(); ++i) {
            c = s.charAt(i);
            switch(c) {
                case '[':
                case ']':
                case '*':
                case '?':
                result.append("[" + c + "]");
                break;
            default:
                result.append(c);
                break;
            }
        }
        
        return result.toString();
    }
    
    public void close() {
        database.close();
    }

   
    public void beginTransaction() {
        database.beginTransaction();
    }

    public void endTransaction() {
        database.endTransaction();
    }

    public void setTransactionSuccessful() {
        database.setTransactionSuccessful();
    }

    public boolean inTransaction() {
        return database.inTransaction();
    }

    
    public void beginSingleTransaction() {
        if (inTransaction())
            ++innerBogusTransactionCount;
        else
            database.beginTransaction();
    }

    public void endSingleTransaction() {
        if (innerBogusTransactionCount > 0) {
            --innerBogusTransactionCount;
            return;
        }
        if (!inTransaction())
            return;
        
        database.endTransaction();
    }

    public void setSingleTransactionSuccessful() {
        if (!inTransaction() || innerBogusTransactionCount > 0)
            return;

        database.setTransactionSuccessful();
    }

    public int getInnerBogusTransactionCount() {
        return innerBogusTransactionCount;
    }

    public void setInnerBogusTransactionCount(int innerBogusTransactionCount) {
        this.innerBogusTransactionCount = innerBogusTransactionCount;
    }

    
    public void execSql(String sql, Object... params) {
        database.execSQL(sql, params);
    }


    public Object[][] execSqlQuery(String sql, Object... params) {
        String[] sParams = new String[params.length];
        for(int i = 0; i < params.length; ++i)
            sParams[i] = params[i].toString();
        
        AbstractWindowedCursor cursor = (AbstractWindowedCursor) database
                .rawQuery(sql, sParams);
        
        try {
            int colCount = cursor.getColumnCount();

            if (colCount == 0)
                return new Object[0][0];

            Object[][] result = new Object[cursor.getCount()][];

            for(int i = 0; i < result.length; ++i) {
                cursor.moveToNext();
                Object[] row = new Object[colCount];
                for(int c = 0; c < colCount; ++c) {
                    row[c] = getObjectForItem(cursor, c);
                }

                result[i] = row;
            }
            
            return result;
        }
        finally {
            cursor.close();
        }
    }


    public Object[] execSqlQuerySingleColumn(String sql, Object... params) {
        Object[][] data = execSqlQuery(sql, params);
        return Utilities.extractColumn(data, 0);
    }
    
    @SuppressWarnings("unchecked")
    public <T> T execSqlQuerySingleItem(String sql, Object... params) {
        String[] sParams = new String[params.length];
        for(int i = 0; i < params.length; ++i)
            sParams[i] = params[i].toString();  // TODO: Blobs?
        
        AbstractWindowedCursor cursor = (AbstractWindowedCursor) database
                .rawQuery(sql, sParams);

        try {
            if (cursor.getColumnCount() == 0 || cursor.getCount() == 0)
                return null;

            cursor.moveToNext();
            return (T) getObjectForItem(cursor, 0);
        }
        finally {
            cursor.close();
        }
    }
    
    /**
     *         Retrieve a value from the settings table
     * @param key
     * @param def
     * @return
     */
    public <T> T getSettingsValue(String key) {
        return execSqlQuerySingleItem("select value from settings where key=?",
                key);
    }
    

    /**
     * Retrieve an integer value from the settings table.
        default -- Default value to return if key was not found
     * @return
     */
    public int getSettingsInt(String key, int def) {
        String value = execSqlQuerySingleItem(
                "select value from settings where key=?", key);

        if (value == null)
            return def;
        
        try {
            return Integer.parseInt(value);
        }
        catch (NumberFormatException e) {
            return def;
        }
    }
    
    

    

}
