package de.mbutscher.wikiandpad.defaultwikilang;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import de.mbutscher.wikiandpad.utils.StringOps;
import de.mbutscher.wikiandpad.utils.Utilities;

class WikiLinkPath implements Cloneable {


    public static class PrefixSilenceAndPageName {
        public String prefix;
        public int silence;
        public String pageName;

        public PrefixSilenceAndPageName(String prefix, int silence,
                String pageName) {
            this.prefix = prefix;
            this.silence = silence;
            this.pageName = pageName;
        }
    }
    
    
    
    protected int upwardCount;
    protected List<String> components;
    
    public WikiLinkPath(String link, String pageName, int upwardCount,
            List<String> components) {
        assert (link == null) || (pageName == null);
        
        if (pageName != null) {
            // Handle wiki word as absolute link
            this.upwardCount = -1;
            this.components = Arrays.asList(pageName.split("/", -1));
            return;
        }
        
        if (link == null) {
            if (components == null)
                components = new ArrayList<String>(0);
            
            this.upwardCount = upwardCount;
            this.components = components;
            return;
        }
        
        if (".".equals(link)) {
            // Link to self
            this.upwardCount = upwardCount;
            this.components = new ArrayList<String>(0);
            return;
        }
        
        if (link.startsWith("//")) {
            this.upwardCount = -1;
            this.components = Arrays.asList(link.substring(2).split("/", -1));
            return;
        }

        if (link.startsWith("/")) {
            this.upwardCount = 0;
            this.components = Arrays.asList(link.substring(1).split("/", -1));
            return;
        }
        
        List<String> comps = Arrays.asList(link.split("/", -1));
        for(int i = 0; i < comps.size(); ++i) {
            if (!"..".equals(comps.get(i))) {
                this.upwardCount = i + 1;
                this.components = comps.subList(i, comps.size());
                return;
            }
        }
        
        this.upwardCount = comps.size();
        this.components = new ArrayList<String>(0);
    }
    
    @Override
    public WikiLinkPath clone() {
        WikiLinkPath result = new WikiLinkPath(null, null, -1, null);
        result.upwardCount = upwardCount;
        result.components = new ArrayList<String>(components);
        
        return result;
    }
    
    
    @Override
    public String toString() {
        return String.format("WikiLinkPath(upwardCount=%i, components=%s)",
                upwardCount, components);
    }
    
    public boolean isAbsolute() {
        return upwardCount == -1;
    }
    
    public void join(WikiLinkPath otherPath) {
        if (otherPath.upwardCount == -1) {
            upwardCount = -1;
            components = new ArrayList<String>(otherPath.components);
            return;
        }
        else if (otherPath.upwardCount == 0) {
            /* ArrayList<String> newComps = new ArrayList<String>(
                    components.size() + otherPath.components.size());
            
            newComps.addAll(components);
            newComps.addAll(otherPath.components);

            components = newComps;
            
            */
            
            components.addAll(otherPath.components);
        }
        else {
            if (otherPath.upwardCount <= components.size()) {
                components = components.subList(0,
                        components.size() - otherPath.upwardCount);
                components.addAll(otherPath.components);
            }
            else {
                // Going back further than self was deep (eliminating
                // more components than self had)

                if (upwardCount == -1) {
                    // Actually an error (going upward after already reaching root)
                    // TODO: Handle as error?
                    components = new ArrayList<String>(otherPath.components);
                }
                else {
                    // Add up upwardCount of other path after subtracting
                    // number of own components because otherPath walked
                    // over them already
                    upwardCount += otherPath.upwardCount - components.size();

                    components = new ArrayList<String>(otherPath.components);
                }
            }
        }
    }
    
    
    public String getLinkCore() {
        String comps = StringOps.strJoin("/", components);
        
        if (upwardCount == -1)
            return "//" + comps;
        else if (upwardCount == 0)
            return "/" + comps;
        else if (upwardCount == 1)
            return comps;
        else
            return StringOps.strJoin("/",
                    Utilities.multiple("..", upwardCount - 1)) + "/" + comps;
    }
    
    public String resolveWikiWord(WikiLinkPath basePath) {
        if (isAbsolute()) {
            // Absolute is checked separately so basePath can be None if
            // this is absolute
            return StringOps.strJoin("/", components);
        }

        WikiLinkPath absPath = basePath.joinTo(this);
        return StringOps.strJoin("/", absPath.components);
    }
    

    
    protected static int _lenAddOne(String s) {
        return s.length() > 0 ? s.length() + 1 : 0;
    }

    
    /**
     *  If using subpages this is used to resolve a link to the right wiki word
        for autocompletion. It returns (prefix, silence, pageName).
        Autocompletion now searches for all wiki words starting with pageName.
        For all found items it removes the first  silence  characters, prepends
        the  prefix  instead and uses the result as suggestion for autocompletion.
        
        If  prefix  is null autocompletion is not possible.

     * @param basePath
     * @return
     */
    public PrefixSilenceAndPageName resolvePrefixSilenceAndWikiWordLink(
            WikiLinkPath basePath){
        if (isAbsolute())
            return new PrefixSilenceAndPageName("//", 0, resolveWikiWord(null));

        assert basePath.isAbsolute();
        
        if (components.isEmpty()) {
            // link path only consists of ".." -> autocompletion not possible
            if (upwardCount == 0) {
                return new PrefixSilenceAndPageName(null, -1,
                        StringOps.strJoin("/", basePath.components));
            }

            return new PrefixSilenceAndPageName(null, -1,
                    StringOps.strJoin("/", basePath.components.subList(
                            0, basePath.components.size() - upwardCount)));
        }

        if (upwardCount == 0) {
            ArrayList<String> comps = new ArrayList<String>(basePath.components);
            comps.addAll(components);
            
            return new PrefixSilenceAndPageName("/",
                    basePath.resolveWikiWord(null).length() + 1,
                    StringOps.strJoin("/", comps));
        }


        if (upwardCount == 1) {
            List<String> basePathWithoutLast = basePath.components.subList(0,
                    basePath.components.size() - 1);
            
            ArrayList<String> comps = new ArrayList<String>(basePathWithoutLast);
            comps.addAll(components);

            return new PrefixSilenceAndPageName("",
                    _lenAddOne(StringOps.strJoin("/", basePathWithoutLast)),
                    StringOps.strJoin("/", comps));
        }

       
        
        List<String> basePathWithoutSome = basePath.components.subList(0,
                basePath.components.size() - upwardCount);

        String prefix = StringOps.strJoin("/",
                Utilities.multiple("..", upwardCount - 1)) + "/";
        int silence = _lenAddOne(StringOps.strJoin("/", basePathWithoutSome));
        
        ArrayList<String> comps = new ArrayList<String>(basePathWithoutSome);
        comps.addAll(components);

        String pageName = StringOps.strJoin("/", comps);

        return new PrefixSilenceAndPageName(prefix, silence, pageName);
    }


    public WikiLinkPath joinTo(WikiLinkPath otherPath) {
        WikiLinkPath result = clone();
        result.join(otherPath);
        return result;
    }
    
    
    public static boolean isAbsoluteLinkCore(String linkCore) {
        return linkCore.startsWith("//");
    }

    
    /** 
     * Create a link to targetAbsPath relative to baseAbsPath.
        If downwardOnly is false, the link may contain parts to go to parents
            or siblings
        in path (in this wiki language, ".." are used for this).
        If downwardOnly is true, the function may return null if a relative
        link can't be constructed.
     * @param targetAbsPath
     * @param baseAbsPath
     * @param downwardOnly (default: true)
     * @return
     */
    public static WikiLinkPath getRelativePathByAbsPaths(
            WikiLinkPath targetAbsPath, WikiLinkPath baseAbsPath,
            boolean downwardOnly) {
        
        assert targetAbsPath.isAbsolute() && baseAbsPath.isAbsolute();

        ArrayList<String> wordPath = new ArrayList<String> (
                targetAbsPath.components);
        ArrayList<String> baseWordPath = new ArrayList<String> (
                baseAbsPath.components);
        
        WikiLinkPath result = new WikiLinkPath(null, null, -1, null);

        if (downwardOnly) {
            if (baseWordPath.size() >= wordPath.size())
                return null;
            if (!baseWordPath.equals(wordPath.subList(0, baseWordPath.size())))
                return null;
            
            result.upwardCount = 0;
            result.components = wordPath.subList(baseWordPath.size(),
                    wordPath.size());
            return result;
            // TODO: test downwardOnly == False
        }
        else {
            // Remove common path elements
            while (wordPath.size() > 0 && baseWordPath.size() > 0 &&
                    wordPath.get(0) == baseWordPath.get(0)) {
                
                wordPath.remove(0);
                baseWordPath.remove(0);
            }
            
            if (baseWordPath.size() == 0) {
                if (wordPath.size() == 0)
                    return null;  // word == baseWord, TODO return u"." or something

                result.upwardCount = 0;
                result.components = wordPath;
                return result;
            }

            result.upwardCount = baseWordPath.size();
            result.components = wordPath;
            return result;
        }

    }
            
    
}


/*

class _WikiLinkPath(object):




    @staticmethod
    def getRelativePathByAbsPaths(targetAbsPath, baseAbsPath,
            downwardOnly=True):
        """
        Create a link to targetAbsPath relative to baseAbsPath.
        If downwardOnly is False, the link may contain parts to go to parents
            or siblings
        in path (in this wiki language, ".." are used for this).
        If downwardOnly is True, the function may return None if a relative
        link can't be constructed.
        """
        assert targetAbsPath.isAbsolute() and baseAbsPath.isAbsolute()

        wordPath = targetAbsPath.components[:]
        baseWordPath = baseAbsPath.components[:]
        
        result = _WikiLinkPath()
        
        if downwardOnly:
            if len(baseWordPath) >= len(wordPath):
                return None
            if baseWordPath != wordPath[:len(baseWordPath)]:
                return None
            
            result.upwardCount = 0
            result.components = wordPath[len(baseWordPath):]
            return result
        # TODO test downwardOnly == False
        else:
            # Remove common path elements
            while len(wordPath) > 0 and len(baseWordPath) > 0 and \
                    wordPath[0] == baseWordPath[0]:
                del wordPath[0]
                del baseWordPath[0]
            
            if len(baseWordPath) == 0:
                if len(wordPath) == 0:
                    return None  # word == baseWord, TODO return u"." or something

                result.upwardCount = 0
                result.components = wordPath
                return result

            result.upwardCount = len(baseWordPath)
            result.components = wordPath
            return result



*/




