package de.mbutscher.wikiandpad.parsing;

import java.util.regex.Pattern;

import de.mbutscher.wikiandpad.utils.NotCurrentThreadException;


/**
 * Forward declaration of an expression to be defined later -
       used for recursive grammars, such as algebraic infix notation.
       When the expression is known, it is assigned to the Forward variable
       using the #assign() method.

 *
 */
public class Forward extends ParseElementEnhance
        implements NecessaryRegexProvider {

    public Forward() {
        super(null);
    }

    public Forward(ParserElement expr) {
        super(expr);
    }
    
    
    @Override
    public ParserElement streamline() {
        if (!streamlined) {
            streamlined = true;
            if (expr != null)
                expr.streamline();
        }
        return this;
    }

    
    @Override
    public boolean getNamedElementNeedsPacking() {
        if (expr == null)
            return false;
        
        return (expr.getNamedElementNeedsPacking() &&
                !expr.hasResultsName() && !hasResultsName());
    }
    
    public Pattern getRegex() {
        if (!(expr instanceof NecessaryRegexProvider))
            return null;
        
        if (buildingRegex)
            return null;
        
        buildingRegex = true;
        try {
            return ((NecessaryRegexProvider) expr).getRegex();
        }
        finally {
            buildingRegex = false;
        }
    }

    public int getRegexFlagsMask() {
        if (!(expr instanceof NecessaryRegexProvider))
            return 0;
        
        if (buildingRegex)
            return -1;
        
        buildingRegex = true;
        try {
            return ((NecessaryRegexProvider) expr).getRegexFlagsMask();
        }
        finally {
            buildingRegex = false;
        }
    }

    public boolean isRegexComplete() {
        if (!(expr instanceof NecessaryRegexProvider))
            return false;
        
        if (buildingRegex)
            return false;

        buildingRegex = true;
        try {
            return ((NecessaryRegexProvider) expr).isRegexComplete();
        }
        finally {
            buildingRegex = false;
        }
    }

    /**
     * We may have a chain of Forward s pointing to other Forward s so find
       the head of this chain.
     * @return
     */
    protected Forward getHeadForward() {
        Forward ret = this;
        while (ret.expr instanceof Forward)
            ret = (Forward) ret.expr;
        
        return ret;
    }
    
    public void assign(ParserElement other) {
        Forward head = getHeadForward();
        head.expr = other;
        head.mayReturnEmpty = other.mayReturnEmpty;
        head.mayIndexError = other.mayIndexError;
    }

    @Override
    protected ParsingResult parseImpl(String instring, int loc,
            ParsingState state, boolean doActions) throws ParseException,
            NotCurrentThreadException {
        
        ParsingResult parsingResult = super.parseImpl(instring, loc, state,
                doActions);
        
        if (!parsingResult.fromSingleNode &&
                parsingResult.retTokens.size() > 0 &&
                expr.getNamedElementNeedsPacking() &&
                !expr.hasResultsName() && hasResultsName()) {

            return new ParsingResult(parsingResult.nextLoc,
                    SyntaxNode.buildSyntaxNode(
                    parsingResult.retTokens, loc, resultsName));
        }
        
        return parsingResult;
    }
    
    @Override
    public String toString() {
        return String.format("%s:(%s)", getClass().getName(), resultsName);
    }

}


/*


def streamline( self ):
    if not self.streamlined:
        self.streamlined = True
        if self.expr is not None:
            self.expr.streamline()
    return self


def __str__( self ):
    if hasattr(self,"name"):
        return self.name

    # TODO: Not thread-safe
    self._revertClass = self.__class__
    self.__class__ = _ForwardNoRecurse
    try:
        if self.expr is not None:
            retString = _ustr(self.expr)
        else:
            retString = "None"
    finally:
        self.__class__ = self._revertClass
    return self.__class__.__name__ + ": " + retString

def copy(self):
    if self.expr is not None:
        return super(Forward,self).copy()
    else:
        ret = Forward()
        ret << self
        return ret


*/
            
