package de.mbutscher.wikiandpad.parsing;

import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.ListIterator;

import android.util.Log;

import de.mbutscher.wikiandpad.utils.IterableIterator;
import de.mbutscher.wikiandpad.utils.StringOps;
import de.mbutscher.wikiandpad.utils.FilteredRoIterator;

public class NonTerminalNode extends SyntaxNode implements List<SyntaxNode> {

    protected List<SyntaxNode> sub;
    
    public NonTerminalNode(List<SyntaxNode> sub, int pos, String name) {
        super(pos, name);
        if (sub != null)
            this.sub = sub;
        else
            this.sub = new ArrayList<SyntaxNode>();
    }

    public NonTerminalNode() {
        this(new ArrayList<SyntaxNode>(), -1, null);
    }

    @Override
    public boolean isTerminal() {
        return false;
    }

    public List<SyntaxNode> getChildren() {
        return sub;
    }
        
    public int getChildrenCount() {
        return sub.size();
    }

    @Override
    public List<SyntaxNode> asList() {
        return sub;
    }

    @Override
    public NonTerminalNode asNonTerminal() {
        return this;
    }

    @Override
    public List<String> asStringList(String sep) {
        ArrayList<String> out = new ArrayList<String>();
        for(SyntaxNode item: sub) {
            if (out.size() > 0 && sep.length() > 0)
                out.add(sep);
            
            out.addAll(item.asStringList());
        }
        return null;
    }

    @Override
    public String getString() {
        StringBuilder result = new StringBuilder();
        for(SyntaxNode sn: sub)
            result.append(sn.getString());
        
        return result.toString();        
    }


    @Override
    public void recalcStrLength() {
        int result = 0;
        for(SyntaxNode sn: sub)
            result += sn.getStrLength();
        
        strLength = result;        
    }

    @Override
    public List<SyntaxNode> findNodesForCharPos(int charPos) {
        // Algorithm taken from Python standard lib bisect module
        int lo = 0;
        int hi = getChildrenCount();
        while (lo < hi) {
            int mid = (lo + hi) / 2;
            if (charPos < sub.get(mid).getPos())
                hi = mid;
            else
                lo = mid + 1;
        }

        int index = lo - 1;

        if (index == -1) {
            // Before first token
            return new ArrayList<SyntaxNode>();
        }

        SyntaxNode node = sub.get(index);

        if (charPos >= (node.pos + node.strLength)) {
            // After last token or outside of any token
            return new ArrayList<SyntaxNode>();
        }

        List<SyntaxNode> result = node.findNodesForCharPos(charPos);
        result.add(node);
        return result;
    }
    
    /**
     * Return first element with given name or return null.
     * @param name
     * @param start
     * @return
     */
    public SyntaxNode findFlatByName(final String name, final int start) {
        for (SyntaxNode node: iterFlatByName(name, start))
            return node;
    
        return null; 
    }

    public SyntaxNode findFlatByName(String name) {
        return findFlatByName(name, 0); 
    }

    @Override
    public SyntaxNode getFirstToken() {
        if (sub == null || sub.size() == 0)
            return null;

        if (sub.get(0) == null)
            Log.d("WikiAndPad", "Error");

        return sub.get(0);
    }

    @Override
    protected void pprintRecurs(int ind, int inc, StringBuilder result) {
        result.append(StringOps.multiply(" ", ind) + String.format(
                "NtNode(%s, %s, '%s', %s ", pos, getStrLength(), name,
                pprintDict()));
    
        result.append("[\n");
        
        int newInd = ind + inc;
        for (SyntaxNode item: sub) {
            item.pprintRecurs(newInd, inc, result);
            result.append("\n");
        }
        
        result.append(StringOps.multiply(" ", ind) + "]  ('" + name + "')");
    }
    
    
    public Iterator<SyntaxNode> iterFlat(final int start) {
        return sub.listIterator(start);
    }
    
    public IterableIterator<SyntaxNode> iterDeep(final int start) {
        
        class SpecialIterator implements IterableIterator<SyntaxNode> {
            SyntaxNode nextResult = null;
            java.util.Stack<Iterator<SyntaxNode>> subIteratorStack;
            
            public SpecialIterator(Iterator<SyntaxNode> subIterator) {
                subIteratorStack = new java.util.Stack<Iterator<SyntaxNode>>();
                subIteratorStack.push(subIterator);
                next();
            }
            
            public boolean hasNext() {
                return subIteratorStack.size() > 0;
            }

            public SyntaxNode next() {
                SyntaxNode result = nextResult;
                if (subIteratorStack.peek().hasNext()) {
                    nextResult = subIteratorStack.peek().next();
                    if (nextResult instanceof NonTerminalNode) {
                        subIteratorStack.push(((NonTerminalNode) nextResult)
                                .iterFlat(0));
                    }
                }
                else {
                    subIteratorStack.pop();
                }

                return result;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
            
            public Iterator<SyntaxNode> iterator() {
                return this;
            }
        };
        
        
        IterableIterator<SyntaxNode> result = new SpecialIterator(iterFlat(start));
        return result;
    }
    
    
    public IterableIterator<SyntaxNode> iterFlatByName(final String name, final int start) {
        FilteredRoIterator<SyntaxNode> result =
                new FilteredRoIterator<SyntaxNode>() {

            @Override
            public boolean filter(SyntaxNode sn) {
                return StringOps.equals(sn.getName(), name);
            }

        };

        result.init(iterFlat(start));
        return result;
    }
    
    public IterableIterator<SyntaxNode> iterDeepByName(final String name, final int start) {
        FilteredRoIterator<SyntaxNode> result =
                new FilteredRoIterator<SyntaxNode>() {

            @Override
            public boolean filter(SyntaxNode sn) {
                return StringOps.equals(sn.getName(), name);
            }

        };

        result.init(iterDeep(start));
        return result;
    }
    
    
    public IterableIterator<SyntaxNode> iterFlatNamed(final int start) {
        FilteredRoIterator<SyntaxNode> result =
                new FilteredRoIterator<SyntaxNode>() {

            @Override
            public boolean filter(SyntaxNode sn) {
                return StringOps.notEmpty(sn.getName());
            }

        };

        result.init(iterFlat(start));
        return result;
    }


    
    


    
    // ---------- Delegates for List interface ----------
    
    public void add(int location, SyntaxNode object) {
        sub.add(location, object);
    }

    public boolean add(SyntaxNode object) {
        return sub.add(object);
    }

    public boolean addAll(Collection<? extends SyntaxNode> arg0) {
        return sub.addAll(arg0);
    }

    public boolean addAll(int arg0, Collection<? extends SyntaxNode> arg1) {
        return sub.addAll(arg0, arg1);
    }

    public void clear() {
        sub.clear();
    }

    public boolean contains(Object object) {
        return sub.contains(object);
    }

    public boolean containsAll(Collection<?> arg0) {
        return sub.containsAll(arg0);
    }

    @Override
    public boolean equals(Object object) {
        return sub.equals(object);
    }

    /**
     * Similar to {@link List#get(int)}, but allows negative indexes like Python.  
     */
    public SyntaxNode get(int location) {
        int s = size();
        
        if (location < 0) {
            location += s;
            if (location < 0)
                location = 0;
        }
        else if (location >= s)
            location = s - 1;

        return sub.get(location);
    }

    @Override
    public int hashCode() {
        return sub.hashCode();
    }

    public int indexOf(Object object) {
        return sub.indexOf(object);
    }

    public boolean isEmpty() {
        return sub.isEmpty();
    }

    public Iterator<SyntaxNode> iterator() {
        return sub.iterator();
    }

    public int lastIndexOf(Object object) {
        return sub.lastIndexOf(object);
    }

    public ListIterator<SyntaxNode> listIterator() {
        return sub.listIterator();
    }

    public ListIterator<SyntaxNode> listIterator(int location) {
        return sub.listIterator(location);
    }

    public SyntaxNode remove(int location) {
        return sub.remove(location);
    }

    public boolean remove(Object object) {
        return sub.remove(object);
    }

    public boolean removeAll(Collection<?> arg0) {
        return sub.removeAll(arg0);
    }

    public boolean retainAll(Collection<?> arg0) {
        return sub.retainAll(arg0);
    }

    public SyntaxNode set(int location, SyntaxNode object) {
        return sub.set(location, object);
    }

    public int size() {
        return sub.size();
    }

    public List<SyntaxNode> subList(int start, int end) {
        return sub.subList(start, end);
    }

    public Object[] toArray() {
        return sub.toArray();
    }

    public <T> T[] toArray(T[] array) {
        return sub.toArray(array);
    }
    
    // ---------- End of Delegates for List interface ----------

    

}
