package de.mbutscher.wikiandpad.parsing;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import android.util.Log;

import de.mbutscher.wikiandpad.utils.NotCurrentThreadException;
import de.mbutscher.wikiandpad.utils.StringOps;
import de.mbutscher.wikiandpad.utils.ThreadStop;

/**
 * Abstract base level parser element class
 *
 */
public abstract class ParserElement implements Cloneable{

    protected List<ParseStartAction> parseStartActions;
    protected List<ValidateAction> validateActions;
    protected List<ParseAction> parseActions;
    
    protected String name;
    protected String errmsg;
    protected String resultsName;
    
    protected boolean mayReturnEmpty; // used when checking for left-recursion
    /** used to optimize exception handling for subclasses that don't advance parse index */
    protected boolean mayIndexError;
    protected boolean streamlined;

    /** To avoid endless recursion */
    protected boolean optimizing = false;
    /** To avoid endless recursion */
    protected boolean buildingRegex = false;

    public String inspect = null;



    public ParserElement() {
        parseStartActions = new ArrayList<ParseStartAction>();
        validateActions = new ArrayList<ValidateAction>();
        parseActions = new ArrayList<ParseAction>();
        
        mayReturnEmpty = false;
        mayIndexError = true;
        streamlined = false;

        // TODO: Much
    }
    
    public ParserElement copy() {
        try {
            ParserElement result = (ParserElement) clone();
            result.parseStartActions = new ArrayList<ParseStartAction>(
                    parseStartActions);
            result.validateActions = new ArrayList<ValidateAction>(
                    validateActions);
            result.parseActions = new ArrayList<ParseAction>(
                    parseActions);
            // TODO: Much

            return result;
        }
        catch (CloneNotSupportedException e) {
            return null;
        }
    }
    
    /**
     * Define name for this expression, for use in debugging.
     * @return this object (for convenience)
     */
    public ParserElement setName(String name) {
        this.name = name;
        this.errmsg = "Expected " + name;
        
        return this;
    }
    
    /**
     * Define name for referencing matching tokens as a nested attribute of the
     * returned parse results. NOTE: this returns a *copy* of the original
     * ParserElement object; this is so that the client can define a basic
     * element, such as an integer, and reference it in multiple places with
     * different names.
     */
    public ParserElement setResultsName(String name) {
        ParserElement result = copy();
        result.resultsName = name;
        
        return result;
    }
    
    /**
     * Define name for referencing matching tokens as a nested attribute of the
     * returned parse results.
     * @return this object (for convenience)
     */
    public ParserElement setResultsNameNoCopy(String name) {
        resultsName = name;
        
        return this;
    }

    public String getResultsName() {
        return resultsName;
    }
    
    public boolean hasResultsName() {
        return resultsName != null && resultsName.length() > 0;
    }
   
    public ParserElement setParseAction(ParseAction pa) {
        parseActions = new ArrayList<ParseAction>(1);
        parseActions.add(pa);
                
        return this;
    }

    public ParserElement addParseAction(ParseAction pa) {
        parseActions.add(pa);
                
        return this;
    }
    
    public ParserElement setValidateAction(ValidateAction va) {
        validateActions = new ArrayList<ValidateAction>(1);
        validateActions.add(va);
                
        return this;
    }
    
    public ParserElement addValidateAction(ValidateAction pa) {
        validateActions.add(pa);
                
        return this;
    }
    public ParserElement setParseStartAction(ParseStartAction psa) {
        parseStartActions = new ArrayList<ParseStartAction>(1);
        parseStartActions.add(psa);
                
        return this;
    }
    
    public ParserElement addParseStartAction(ParseStartAction psa) {
        parseStartActions.add(psa);
                
        return this;
    }
    
    /**
     * If the tokens returned by {@link #postParse(String, int, ParsingState, List)}
     * are a list and this element has a resultsName, the list is packed
     * into a named NonTerminalSyntaxNode before
     * further processing if this returns True
     */
    public boolean getNamedElementNeedsPacking() {
        return false;
    }

    protected abstract ParsingResult parseImpl(String instring, int loc,
            ParsingState state, boolean doActions) throws ParseException,
            NotCurrentThreadException;
    
    public ParsingResult postParse(String instring, int loc,
            ParsingState state, ParsingResult parsingResult) {
        return parsingResult;
    }
    
    @SuppressWarnings("null")
    public ParsingResult parse(String instring, int loc, ParsingState state,
            boolean doActions) throws ParseException, NotCurrentThreadException {

        boolean validResultName = resultsName != null && !"".equals(resultsName);
        boolean debugging = false;

        if (validResultName)
            state.nameStack.push(resultsName);
        
        state.dictStack.push(resultsName).put("parserElement", this);
        try {
            for(ParseStartAction psa: parseStartActions) {
                psa.parseStartAction(instring, loc, state, this);
            }

            ParsingResult parsingResult;
            int tokensStart;

            if (debugging) {
                parsingResult = null;
                        // TODO:
/*
                #~ print ("Match",self,"at loc",loc,"(%d,%d)" % ( lineno(loc,instring), col(loc,instring) ))
                if (self.debugActions[0] ):
                    self.debugActions[0]( instring, loc, self )
                if callPreParse and self.callPreparse:
                    preloc = self.preParse( instring, loc, state )
                else:
                    preloc = loc
*/
                tokensStart = loc;
/*
                try:
                    try:
                        loc,tokens = self.parseImpl( instring, preloc, state, doActions )
                    except IndexError:
                        raise ParseException( instring, len(instring), self.errmsg, self )
                except ParseBaseException, err:
                    #~ print ("Exception raised:", err)
                    if self.debugActions[2]:
                        self.debugActions[2]( instring, tokensStart, self, err )
                    if self.failAction:
                        self.failAction( instring, tokensStart, self, err )
                    raise
 */
            }
            else {
                tokensStart = loc;
                
                if (inspect != null)
                    Log.d("WikiAndPad", "Before parse " + inspect);

                try {
                    if (mayIndexError || loc >= instring.length()) {
                        try {
                            parsingResult = parseImpl(instring, loc, state,
                                    doActions);
                        }
                        catch(IndexOutOfBoundsException e) {
                            if (inspect != null)
                                Log.d("WikiAndPad", "OutOfBounds " + inspect);
                            throw new ParseException(instring,
                                    instring.length(), errmsg, this);
                        }
                    }
                    else {
                        parsingResult = parseImpl(instring, loc, state,
                                doActions);
                    }
                    if (inspect != null)
                        Log.d("WikiAndPad", "Parsed " + inspect);
                }
                catch(ParseException e)
                {
                    if (inspect != null) {
                        Log.d("WikiAndPad", "ParseException " + inspect + ":" +
                                    Integer.toString(e.getLoc()) + ":" +
                                    e.getMsg());
                    }
                    throw e;
                }
            }
            
            if (parsingResult.fromSingleNode) {
                parsingResult.retTokens.set(0, SyntaxNode.buildSyntaxNode(
                        parsingResult.retTokens.get(0), tokensStart,
                        resultsName));
                parsingResult.fromSingleNode = false;
            }

            
            parsingResult = postParse( instring, tokensStart, state,
                    parsingResult);

            loc = parsingResult.nextLoc;

            SyntaxNode retTokens;
            
            if (parsingResult.fromSingleNode) {
                List<SyntaxNode> tokens = Arrays.asList(new SyntaxNode[] {
                        SyntaxNode.buildSyntaxNode(
                        parsingResult.retTokens.get(0), tokensStart,
                        resultsName)
                });
                retTokens = SyntaxNode.buildSyntaxNode(tokens, tokensStart);
            }
            else {
                if (getNamedElementNeedsPacking() && resultsName != null &&
                        !"".equals(resultsName)) {
                    retTokens = SyntaxNode.buildSyntaxNode(
                            parsingResult.retTokens, tokensStart, resultsName);
                }
                else {
                    retTokens = SyntaxNode.buildSyntaxNode(
                            parsingResult.retTokens, tokensStart);
                }
            }


            /* SyntaxNode nodeToValidate;
            if (getNamedElementNeedsPacking() && resultsName != null &&
                    !"".equals(resultsName))
                nodeToValidate = SyntaxNode.buildSyntaxNode(parsingResult.retTokens,
                        tokensStart, resultsName);
            else
                nodeToValidate = SyntaxNode.buildSyntaxNode(parsingResult.retTokens,
                        tokensStart); */
                
            for(ValidateAction psa: validateActions) {
                psa.validateAction(instring, tokensStart, state, retTokens);
            }

            if (inspect != null)
                Log.d("WikiAndPad", "Validated " + inspect);
            
            /*
            except ParseBaseException, err:
                #~ print "Exception raised in user validate action:", err
                if (self.debugActions[2] ):
                    self.debugActions[2]( instring, tokensStart, self, err )
                raise
            */
            
            // ParseProcessingData ppd = new ParseProcessingData(retTokens,
            //         tokensStart);
            
            //ParseProcessingData ppd = new ParseProcessingData(parsingResult.retTokens,
            //        tokensStart);
            
            // ppd.fromSingleNode = false;  //???
            
            if (!parseActions.isEmpty() && (doActions /* or self.callDuringTry */)) {
                if (debugging) {
                    // TODO:
            /*
                    try:
                        for fn in self.parseAction:
                            tokens = fn( instring, tokensStart, state, retTokens )
                            if tokens is not None:
                                retTokens = tokens
                    except ParseBaseException, err:
                        #~ print "Exception raised in user parse action:", err
                        if (self.debugActions[2] ):
                            self.debugActions[2]( instring, tokensStart, self, err )
                        raise
        
             */
                }
                else {
                    SyntaxNode newRetTokens;
                    
                    for(ParseAction pa: parseActions) {
                        newRetTokens = pa.parseAction(instring,
                                tokensStart, state, retTokens);
                        
                        if (newRetTokens != null)
                            retTokens = newRetTokens;
                    }
                }
            }

            if (inspect != null)
                Log.d("WikiAndPad", "Actions done " + inspect);

            if (debugging) {
                // TODO:
                /*
                    #~ print ("Matched",self,"->",retTokens.asList())
                    if (self.debugActions[1] ):
                        self.debugActions[1]( instring, tokensStart, loc, self, retTokens )
                */
            }
            
            /*
            if isinstance(retTokens, SyntaxNode):
                if retTokens.name:
                    retTokens = [retTokens]
                else:
                    retTokens = retTokens.asList()
            */

            /*ppd.encapsulateIfNeeded();

            if (ppd.fromSingleNode &&
                    StringOps.notEmpty(ppd.getFirstToken().name)) {
                ppd.fromSingleNode = false;
            }*/


            ParsingResult result;

            if (StringOps.notEmpty(retTokens.name))
                result = new ParsingResult(loc, retTokens);
            else
                result = new ParsingResult(loc, retTokens.asList());

            state.threadstop.testValidThread();
            return result;
        }
        finally {
            state.dictStack.pop();
            if (validResultName)
                state.nameStack.pop();
        }
    }
    

    public ParsingResult parse(String instring, int loc, ParsingState state)
            throws ParseException, NotCurrentThreadException {
        return parse(instring, loc, state, true);
    }


    public ParsingResult tryParse(String instring, int loc, ParsingState state)
            throws ParseException, NotCurrentThreadException {
        return parse(instring, loc, state, false);
    }
    
    /**
     * Returns a ParsingState containing namestack, fullstack, instring and
     * revInstring. The namestack contains strings with all named tokens.
     * fullstack is a StackedCopyDict instring is the complete text to parse
     * revInstring is the reserved instring
     * 
     * @param instring
     * @param baseDict
     * @param threadstop
     * @return
     */
    public ParsingState buildStartState(String instring,
            Map<String, Object> baseDict, ThreadStop threadstop) {
        return new ParsingState(instring, baseDict, threadstop);
    }
    
    
    /**
     * Execute the parse expression with the given string. This is the main
     * interface to the client code, once the complete expression has been
     * built.
     * 
     * If you want the grammar to require that the entire input string be
     * successfully parsed, then set parseAll to True (equivalent to ending the
     * grammar with StringEnd()).
     * 
     * Note: parseString implicitly calls expandtabs() on the input string, in
     * order to report proper column numbers in parse actions. If the input
     * string contains tabs and the grammar uses parse actions that use the loc
     * argument to index into the string being parsed, you can ensure you have a
     * consistent view of the input string by: <br>
     * - calling parseWithTabs on your
     * grammar before calling parseString (see
     * L{I{parseWithTabs}<parseWithTabs>}) <br>
     * - define your parse action using the full (s,loc,toks) signature, and
     * reference the input string using the parse action's s argument <br>
     * - explictly expand the tabs in your input string before
     * calling parseString
     * 
     * @param instring
     * @param parseAll
     * @param baseDict
     * @param threadstop
     * @return
     * @throws NotCurrentThreadException 
     * @throws ParseException 
     */
    public List<SyntaxNode> parseString(String instring, boolean parseAll,
            Map<String, Object> baseDict, ThreadStop threadstop)
            throws ParseException, NotCurrentThreadException {
        if (!streamlined)
            streamline();
        
        ParsingState state = buildStartState(instring, baseDict, threadstop);
        ParsingResult parsingResult = parse(instring, 0, state);
        if (parseAll) {
            new StringEnd().parse(instring, parsingResult.nextLoc, state);
        }
        return parsingResult.retTokens;
    }
    
    public And and(ParserElement other) {
        return new And(new ParserElement[] {this, other});
    }

    public MatchFirst or(ParserElement other) {
        return new MatchFirst(new ParserElement[] {this, other});
    }

    public ParserElement streamline() {
        streamlined = true;
        return this;
    }

    public ParserElement optimize(OptimizerOptions options)  {
        ParserElement result = streamline();
        return result._realOptimize(options);
    }
    
    protected void _optimizeSub(OptimizerOptions options) {
    }
    
    protected ParserElement _optimizeSelf(OptimizerOptions options) {
        return this;
    }
    
    protected ParserElement _realOptimize(OptimizerOptions options) {
        if (optimizing)
            return this;
        optimizing = true;  // Prevent infinite recursion
        _optimizeSub(options);
        return _optimizeSelf(options);
    }

    public void checkRecursion(List<ParserElement> parseElementList) {
    }

    /**
     * Dummy function, tabs are parsed always.
     * @return this object (for convenience)
     */
    public ParserElement parseWithTabs() {
        return this;
    }
    
    @Override
    public String toString() {
        return name;
    }
    
}
