package de.mbutscher.wikiandpad.parsing;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

/**
 *  A stacked dictionary is technically a stack of dictionaries plus one base
    dictionary which is at the bottom of the stack and can't be dropped off.
    Along with get and set operations which are always executed on the
    stack top dictionary there are functions to push() a new dictionary on
    the stack (which is automatically a copy of the previous stack top) and
    to pop() the stack top dict.

 * @author Michael Butscher
 *
 */
public class StackedCopyDict<K, V> implements StackedDict<K, V> {

    protected Map<K, V> baseDict;
    protected Stack<Map<K, V>> dictStack;
    protected String baseName;
    protected Stack<String> nameStack;


    public StackedCopyDict(Map<K, V> baseDict, String baseName) {
        if (baseDict == null)
            this.baseDict = new HashMap<K, V>();
        else
            this.baseDict = baseDict;
        
        dictStack = new Stack<Map<K, V>>();
        this.baseName = baseName;
        nameStack = new Stack<String>();
    }
    
    public StackedCopyDict(Map<K, V> baseDict) {
        this(baseDict, null);
    }
    


    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#getTopDict()
     */
    public Map<K, V> getTopDict() {
        if (dictStack.size() > 0)
            return dictStack.peek();
        else
            return baseDict;
    }
    
    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#peek()
     */
    public Map<K, V> peek() {
        return getTopDict();
    }


    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#getSubTopDict()
     */
    public Map<K, V> getSubTopDict() {
        if (dictStack.size() > 1)
            return dictStack.get(dictStack.size() - 2);
        else
            return baseDict;
    }
    
    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#getNamedDict(java.lang.String)
     */
    public Map<K, V> getNamedDict(String name) {
        for(int i = nameStack.size() - 1; i > -1; --i) {
            if (name.equals(nameStack.get(i)))
                return dictStack.get(i);
        }
        
        if (name.equals(baseName))
            return baseDict;
        
        return null;
    }
    
    @Override
    public String toString() {
        return "<StackedCopyDict " + getTopDict().toString() + ">";
    }
    
    
    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#push(java.lang.String, java.util.Map)
     */
    public Map<K, V> push(String newName, Map<K, V> newDict) {
        if (newDict == null)
            newDict = new HashMap<K, V>(getTopDict());
        
        dictStack.push(newDict);
        nameStack.push(newName);
        
        return getTopDict();
    }
    
    public Map<K, V> push(String newName) {
        return push(newName, null);
    }

    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#pop()
     */
    public Map<K, V> pop() {
        nameStack.pop();
        return dictStack.pop();
    }

    // Implement parts of Map interface
    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#containsKey(java.lang.Object)
     */
    public boolean containsKey(Object key) {
        return getTopDict().containsKey(key);
    }


    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#get(java.lang.Object)
     */
    public V get(Object key) {
        return getTopDict().get(key);
    }


    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#put(K, V)
     */
    public V put(K key, V value) {
        return getTopDict().put(key, value);
    }


    /* (non-Javadoc)
     * @see de.mbutscher.wikiandpad.parsing.StackedDict#remove(java.lang.Object)
     */
    public V remove(Object key) {
        return getTopDict().remove(key);
    }
    
}
