package de.mbutscher.wikiandpad.utils;

import java.lang.ref.WeakReference;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;

public class RichMEListener implements ExtendedMiscEventListener {

    protected IdentityWeakArrayList<Object> sources;
    protected ArrayList<String> eventKeys;
    protected boolean enabled;
    
    protected Object listenerKey;
    protected MiscEvent holdingEvent;
    
    protected WeakReference<Object> targetRef;
    protected Method targetMethod;
    
    public RichMEListener(Object listenerKey) {
        if (listenerKey == null)
            listenerKey = this;

        this.enabled = true;
        this.listenerKey = listenerKey;
        this.sources = null;
        this.eventKeys = null;
    }

    @SuppressWarnings("unchecked")
    public RichMEListener deepCopy() {
        RichMEListener result = new RichMEListener(listenerKey);
        result.sources = (IdentityWeakArrayList<Object>) sources.clone();
        result.eventKeys = (ArrayList<String>) eventKeys.clone();

        result.enabled = enabled;
        result.holdingEvent = holdingEvent;
        result.targetRef = targetRef;
        result.targetMethod = targetMethod;

        return result;
    }
    
    
    public RichMEListener addSource(Object source) {
        if (sources == null)
            sources = new IdentityWeakArrayList<Object>();

        sources.add(source);
        return this;
    }
    
    public RichMEListener addEventKey(String key) {
        if (eventKeys == null)
            eventKeys = new ArrayList<String>();

        eventKeys.add(key);
        return this;
    }
    
    public RichMEListener setTarget(MiscEventListener target) {
        targetRef = new WeakReference<Object>(target);
        targetMethod = null;
        return this;
    }
    
    public RichMEListener setTarget(Object target, String methodName) {
        targetRef = new WeakReference<Object>(target);
        targetMethod = null;
        if (setMethod(target, methodName) == null) {
            invalidate();
            throw new IllegalArgumentException("No method '" + methodName +
                    "' with appropriate parameters found");
        }
        return this;
    }
    
    public void subscribeTo(MiscEvent holdingEvent) {
        holdingEvent.addListener(this);
    }
    
    public synchronized void invalidate() {
        targetRef = null;
        targetMethod = null;
    }
    

    protected static final Class<?>[] TARGET_PARAMETER_CLASSES = new Class[] {
            MiscEvent.class, RichMEListener.class
    };
    
    protected static boolean checkMethodSignature(Method m) {
        Class<?>[] parameters = m.getParameterTypes();
        
        if (parameters.length > 2)
            return false;
        
        for(int i = 0; i < parameters.length; ++i) {
            if (!TARGET_PARAMETER_CLASSES[i].isAssignableFrom(parameters[i]))
                return false;
        }
        
        return true;
    }
    
    protected Method setMethod(Object target, String methodName) {
        if (methodName == null) {
            targetMethod = null;
            return null;
        }

        if (target == null) {
            invalidate();
            return null;
        }

        for(Method m: target.getClass().getMethods()) {
            if (methodName.equals(m.getName()) && checkMethodSignature(m)) {
                targetMethod = m;
                return m;
            }
        }
        
        /*
        for(Method m: target.getClass().getDeclaredMethods()) {
            if (methodName.equals(m.getName()) && checkMethodSignature(m)) {
                targetMethod = m;
                return m;
            }
        }
        */
        
        targetMethod = null;
        return null;
    }
    
    public void callTarget(MiscEvent event) {
        Object target = targetRef.get();
        if (target == null) {
            invalidate();
            return;
        }
        
        if (targetMethod != null) {
            try {
                switch (targetMethod.getParameterTypes().length) {
                case 0:
                    targetMethod.invoke(target);
                    break;
                case 1:
                    targetMethod.invoke(target, event);
                    break;
                case 2:
                    targetMethod.invoke(target, event, this);
                    break;
                }
            }
            catch (IllegalArgumentException e) {
                e.printStackTrace();
            }
            catch (IllegalAccessException e) {
                e.printStackTrace();
            }
            catch (InvocationTargetException e) {
                e.printStackTrace();
            }
        }
        else {
            try {
                MiscEventListener l = (MiscEventListener) target;
                l.miscEventHappened(event);
            }
            catch (ClassCastException e) {
                e.printStackTrace();
            }
        }
    }
    
    public boolean matchesEvent(MiscEvent event) {
        if (sources != null && !sources.contains(event.getSource()))
            return false;

        if (eventKeys != null && !event.hasAnyKey(eventKeys))
            return false;
        
        return true;
    }

    
    public void miscEventHappened(MiscEvent event) {
        if (!enabled)
            return;

        if (matchesEvent(event))
            callTarget(event);
    }

    public void setHoldingEvent(MiscEvent holdingEvent) {
        this.holdingEvent = holdingEvent;
    }

    public synchronized boolean isValid() {
        if (targetRef == null)
            return false;
        
        if (targetRef.get() == null) {
            invalidate();
            return false;
        }

        return true;
    }

    public boolean matchesListenerKey(Object listenerKey) {
        return this.listenerKey == listenerKey;
    }

    public void setEnabled(boolean enabled) {
        this.enabled = enabled;
    }
    
    public boolean getEnabled() {
        return enabled;
    }

}
