/*
 * Copyright (C) 2006 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// New class based on ArrayAdapter


package de.mbutscher.wikiandpad.utils.ui;

import java.util.List;

import android.widget.TextView;
import android.content.Context;
import android.util.Log;
import android.view.View;
import android.view.ViewGroup;

import de.mbutscher.wikiandpad.utils.Stringifier;

/*
 * A concrete BaseAdapter that is backed by an array of arbitrary
 * objects.  By default this class expects that the provided resource id references
 * a single TextView.  If you want to use a more complex layout, use the constructors that
 * also takes a field id.  That field id should reference a TextView in the larger layout
 * resource.
 *
 * <p>However the TextView is referenced, it will be filled with the toString() of each object in
 * the array. You can add lists or arrays of custom objects. Override the toString() method
 * of your objects to determine what text will be displayed for the item in the list.
 *
 * <p>To use something other than TextViews for the array display, for instance, ImageViews,
 * or to have some of data besides toString() results fill the views,
 * override {@link #getView(int, View, ViewGroup)} to return the type of view you want.
 */

public class StringArrayListAdapter<T> extends ArrayListAdapter<T> {

    protected Stringifier<T> stringifier;

    /**
     * {@inheritDoc}
     */
    public StringArrayListAdapter(Context context, int resource,
            int textViewResourceId, List<T> objects) {
        super(context, resource, textViewResourceId, objects);
    }

    /**
     * {@inheritDoc}
     */
    public StringArrayListAdapter(Context context, int resource,
            int textViewResourceId, T[] objects) {
        super(context, resource, textViewResourceId, objects);
    }

    /**
     * {@inheritDoc}
     */
    public StringArrayListAdapter(Context context, int resource,
            int textViewResourceId) {
        super(context, resource, textViewResourceId);
    }

    /**
     * {@inheritDoc}
     */
    public StringArrayListAdapter(Context context, int textViewResourceId,
            List<T> objects) {
        super(context, textViewResourceId, objects);
    }

    /**
     * {@inheritDoc}
     */
    public StringArrayListAdapter(Context context, int textViewResourceId,
            T[] objects) {
        super(context, textViewResourceId, objects);
    }


    /**
     * {@inheritDoc}
     */
    public StringArrayListAdapter(Context context, int textViewResourceId) {
        super(context, textViewResourceId);
    }

    public Stringifier<T> getStringifier() {
        return stringifier;
    }

    public void setStringifier(Stringifier<T> stringifier) {
        if (stringifier == null)
            stringifier = new Stringifier.Default<T>();

        this.stringifier = stringifier;
    }

    @Override
    protected void init(Context context, int resource, int textViewResourceId,
            List<T> objects) {
        super.init(context, resource, textViewResourceId, objects);
        
        stringifier = new Stringifier.Default<T>();
    }

    @Override
    protected View createViewFromResource(int position, View convertView, ViewGroup parent,
            int resource) {
        View view;
        TextView text;

        if (convertView == null) {
            view = mInflater.inflate(resource, parent, false);
        } else {
            view = convertView;
        }

        try {
            if (mFieldId == 0) {
                //  If no custom field is assigned, assume the whole resource is a TextView
                text = (TextView) view;
            } else {
                //  Otherwise, find the TextView field within the layout
                text = (TextView) view.findViewById(mFieldId);
            }
        } catch (ClassCastException e) {
            Log.e("ArrayAdapter", "You must supply a resource ID for a TextView");
            throw new IllegalStateException(
                    "ArrayAdapter requires the resource ID to be a TextView", e);
        }

        T item = getItem(position);
        if (item instanceof CharSequence) {
            text.setText((CharSequence)item);
        } else {
            text.setText(stringifier.toString(item));
        }

        return view;
    }

}